package jalview.ws;

import java.util.Enumeration;
import java.util.Vector;

import org.biojava.dasobert.das2.Das2Source;
import org.biojava.dasobert.dasregistry.Das1Source;
import org.biojava.dasobert.dasregistry.DasCoordinateSystem;
import org.biojava.dasobert.dasregistry.DasSource;

import jalview.datamodel.Alignment;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.DBRefSource;
import jalview.datamodel.SequenceI;
import jalview.ws.seqfetcher.ASequenceFetcher;
import jalview.ws.seqfetcher.DbSourceProxy;

/**
 * This is the the concrete implementation of the sequence retrieval interface
 * and abstract class in jalview.ws.seqfetcher. This implements the run-time
 * discovery of sequence database clients, and provides a hardwired main for testing all registered handlers. 
 * 
 */
public class SequenceFetcher extends ASequenceFetcher
{
  /**
   * Thread safe construction of database proxies TODO: extend to a configurable
   * database plugin mechanism where classes are instantiated by reflection and
   * queried for their DbRefSource and version association.
   * 
   */
  public SequenceFetcher()
  {
    addDBRefSourceImpl(jalview.ws.dbsources.EmblSource.class);
    addDBRefSourceImpl(jalview.ws.dbsources.EmblCdsSouce.class);
    addDBRefSourceImpl(jalview.ws.dbsources.Uniprot.class);
    addDBRefSourceImpl(jalview.ws.dbsources.UnprotName.class);
    addDBRefSourceImpl(jalview.ws.dbsources.Pdb.class);
    addDBRefSourceImpl(jalview.ws.dbsources.Pfam.class);
    registerDasSequenceSources();
  }
  /**
   * simple run method to test dbsources.
   * @param argv
   */
  public static void main(String[] argv)
  {
    AlignmentI ds = null;
    Vector noProds = new Vector();
    String usage = "SequenceFetcher.main [<DBNAME> <ACCNO>]\n"
      +"With no arguments, all DbSources will be queried with their test Accession number.\n"
      +"If given two arguments, SequenceFetcher will try to find the DbFetcher corresponding to <DBNAME> and retrieve <ACCNO> from it.";
    if (argv != null && argv.length > 0)
    {
      DbSourceProxy sp = new SequenceFetcher().getSourceProxy(argv[0]);
      if (sp!=null)
      {
        AlignmentI al = null;
        try
        {
          al = sp.getSequenceRecords(argv[1]);
        } catch (Exception e)
        {
          e.printStackTrace();
          System.err.println("Error when retrieving "+argv[1]+" from "+argv[0]+"\nUsage: "+usage);
        }
        SequenceI[] prod = al.getSequencesArray();
        if (al!=null)
        {
          for (int p = 0; p < prod.length; p++)
          {
            System.out.println("Prod " + p + ": "
                    + prod[p].getDisplayId(true) + " : "+prod[p].getDescription());
          }
        }
        return;
      } else {
        System.err.println("Can't resolve "+argv[0]+" as a database name. Allowed values are :\n"+new SequenceFetcher().getSupportedDb());
      }
      System.out
              .println(usage);
    }
    ASequenceFetcher sfetcher = new SequenceFetcher();
    String[] dbSources = sfetcher.getSupportedDb();
    for (int dbsource=0; dbsource<dbSources.length;dbsource++)
    {
      String db = dbSources[dbsource];
      // skip me
      if (db.equals(DBRefSource.PDB))
        continue;
      DbSourceProxy sp = sfetcher.getSourceProxy(db);
      System.out
              .println("Source: "+sp.getDbName()+" (" + db + "): retrieving test:" + sp.getTestQuery());
      AlignmentI al = null;
      try
      {
        al = sp.getSequenceRecords(sp.getTestQuery());
        if (al != null && al.getHeight() > 0 && sp.getDbSourceProperties()!=null)
        {
          boolean dna = sp.getDbSourceProperties().containsKey(
                  DBRefSource.DNACODINGSEQDB)
                  || sp.getDbSourceProperties().containsKey(
                          DBRefSource.DNASEQDB)
                  || sp.getDbSourceProperties().containsKey(
                          DBRefSource.CODINGSEQDB);
          // try and find products
          String types[] = jalview.analysis.CrossRef.findSequenceXrefTypes(
                  dna, al.getSequencesArray());
          if (types != null)
          {
            System.out.println("Xref Types for: " + (dna ? "dna" : "prot"));
            for (int t = 0; t < types.length; t++)
            {
              System.out.println("Type: " + types[t]);
              SequenceI[] prod = jalview.analysis.CrossRef
                      .findXrefSequences(al.getSequencesArray(), dna,
                              types[t]).getSequencesArray();
              System.out.println("Found "
                      + ((prod == null) ? "no" : "" + prod.length)
                      + " products");
              if (prod != null)
              {
                for (int p = 0; p < prod.length; p++)
                {
                  System.out.println("Prod " + p + ": "
                          + prod[p].getDisplayId(true));
                }
              }
            }
          }
          else
          {
            noProds.addElement((dna ? new Object[]
            { al, al } : new Object[]
            { al }));
          }

        }
      } catch (Exception ex)
      {
        System.out.println("ERROR:Failed to retrieve test query.");
        ex.printStackTrace(System.out);
      }
      if (al == null)
      {
        System.out.println("ERROR:No alignment retrieved.");
        StringBuffer raw = sp.getRawRecords();
        if (raw != null)
          System.out.println(raw.toString());
        else
          System.out.println("ERROR:No Raw results.");
      }
      else
      {
        System.out.println("Retrieved " + al.getHeight() + " sequences.");
        for (int s = 0; s < al.getHeight(); s++)
        {
          SequenceI sq = al.getSequenceAt(s);
          while (sq.getDatasetSequence() != null)
          {
            sq = sq.getDatasetSequence();

          }
          if (ds == null)
          {
            ds = new Alignment(new SequenceI[]
            { sq });

          }
          else
          {
            ds.addSequence(sq);
          }
        }
      }
      System.out.flush();
      System.err.flush();

    }
    if (noProds.size() > 0)
    {
      Enumeration ts = noProds.elements();
      while (ts.hasMoreElements())

      {
        Object[] typeSq = (Object[]) ts.nextElement();
        boolean dna = (typeSq.length > 1);
        AlignmentI al = (AlignmentI) typeSq[0];
        System.out.println("Trying getProducts for "
                + al.getSequenceAt(0).getDisplayId(true));
        System.out.println("Search DS Xref for: " + (dna ? "dna" : "prot"));
        // have a bash at finding the products amongst all the retrieved
        // sequences.
        SequenceI[] seqs = al.getSequencesArray();
        Alignment prodal = jalview.analysis.CrossRef.findXrefSequences(
                seqs, dna, null, ds);
        System.out.println("Found "
                + ((prodal == null) ? "no" : "" + prodal.getHeight()) + " products");
        if (prodal != null)
        {
          SequenceI[] prod = prodal.getSequencesArray(); // note
          // should
          // test
          // rather
          // than
          // throw
          // away
          // codon
          // mapping
          // (if
          // present)
            for (int p = 0; p < prod.length; p++)
          {
            System.out.println("Prod " + p + ": "
                    + prod[p].getDisplayId(true));
          }
        }
      }

    }
  }
  /**
   * query the currently defined DAS source registry for sequence sources and add a DasSequenceSource instance for each source to the SequenceFetcher source list.
   */
  public void registerDasSequenceSources() {
    DasSource[] sources = jalview.ws.DasSequenceFeatureFetcher.getDASSources();
    for (int s=0;s<sources.length; s++)
    {
      Das1Source d1s=null;
      if (sources[s].hasCapability("sequence"))
      {
        if (sources[s] instanceof Das2Source)
        {
          if (((Das2Source)sources[s]).hasDas1Capabilities()) {
            try {
              d1s = org.biojava.dasobert.das2.DasSourceConverter.toDas1Source((Das2Source) sources[s]);
            } catch (Exception e)
            {
              System.err.println("Ignoring DAS2 sequence source "+sources[s].getNickname()+" - couldn't map to Das1Source.\n");
              e.printStackTrace();
            }
          }
        } else {
          if (sources[s] instanceof Das1Source)
          { 
            d1s = (Das1Source) sources[s];
          }
        }
      }
      if (d1s!=null)
      {
        DasCoordinateSystem[] css = d1s.getCoordinateSystem();
        for (int c=0;c<css.length;c++)
        {
          try {
            addDbRefSourceImpl(
                    new jalview.ws.dbsources.DasSequenceSource("das:"+d1s.getNickname()+" ("+css[c].getName()+")",
                            css[c].getName(),d1s, css[c]));
          } catch (Exception e)
          {
            System.err.println("Ignoring sequence coord system "+c+" ("+
                    css[c].getName()+") for source "+d1s.getNickname()
                    + "- threw exception when constructing fetcher.\n");
            e.printStackTrace();
          }
        }
      }
          
    }
  }

}
